import { system, world, ItemStack, BlockPermutation } from "@minecraft/server";
let netherStarLock = false;
const oreList = [
    "minecraft:coal_ore", "minecraft:iron_ore", "minecraft:gold_ore", "minecraft:diamond_ore", "minecraft:emerald_ore", "minecraft:lapis_ore", "minecraft:redstone_ore", "minecraft:lit_redstone_ore", "minecraft:copper_ore",
    "minecraft:deepslate_coal_ore", "minecraft:deepslate_iron_ore", "minecraft:deepslate_gold_ore", "minecraft:deepslate_diamond_ore", "minecraft:deepslate_emerald_ore", "minecraft:deepslate_lapis_ore", "minecraft:deepslate_redstone_ore", "minecraft:lit_deepslate_redstone_ore", "minecraft:deepslate_copper_ore",
    "minecraft:quartz_ore", "minecraft:nether_gold_ore", "minecraft:ancient_debris", 
];
const oreMaxDrop = {
    "redstone": {},
    "lapis": {},
    "copper": {},
    "quartz": {},
    "nether_gold": {}
}
const oreDropList = {
    "coal": "minecraft:coal",
    "iron": "minecraft:iron_ingot",
    "gold": "minecraft:gold_ingot",
    "diamond": "minecraft:diamond",
    "emerald": "minecraft:emerald",
    "lapis": "minecraft:lapis_lazuli",
    "redstone": "minecraft:redstone",
    "copper": "minecraft:copper_ingot",
    "quartz": "minecraft:quartz",
    "nether_gold": "minecraft:gold_nugget",
    "ancient_debris": "minecraft:netherite_scrap",
}
const conversionDirtListHoe = {
    "podzol": "dirt",
    "mycelium": "dirt",
    "dirt": "farmland",
    "grass_block": "farmland",
    "coarse_dirt": "dirt",
    "grass_path": "dirt",
    "farmland": "farmland",
}
const conversionDirtListShovel = {
    "podzol": {},
    "dirt_with_roots": {},
    "mycelium": {},
    "dirt": {},
    "grass_block": {},
    "coarse_dirt": {},
    "farmland": {},
}
const woodConversionList = {
    "oak": {},
    "spruce": {},
    "birch": {},
    "jungle": {},
    "acacia": {},
    "dark_oak": {},
    "crimson": {},
    "warped": {},
    "mangrove": {},
    "cherry": {},
}
const dirToMethod = {
    "north": "north",
    "south": "south",
    "east": "east",
    "west": "west",
    "up": "above",
    "down": "below",
}
const oppoDir = {
    "north": "south",
    "south": "north",
    "east": "west",
    "west": "east",
    "up": "down",
    "down": "up",
}
const stateGrowth = {
    "small": "medium",
    "medium": "large",
}
const shardAmount = {
    "small": 1,
    "medium": 2,
    "large": 3,
}
const vixiumRaw = {
    "small": 0,
    "medium": 1,
    "large": 2,
}
const entityConversionList = {
    "nether_star_entity": "minecraft:nether_star",
    "vixium_hoe_throwing": "vi:vixium_hoe",
    "vixium_sword_throwing": "vi:vixium_sword",
    "vixium_shovel_throwing": "vi:vixium_shovel",
}
world.gameRules.sendCommandFeedback = false;

function damageEntity(main, player, amount) {
    let returnSignal = false;
    main.dimension.getEntities({location: main.location, maxDistance: 3}).forEach(entity => {
        let family = entity.getComponent("minecraft:type_family");
        if (family?.hasTypeFamily("monster") || family?.hasTypeFamily("player") && entity !== player) {
            entity.applyDamage(amount, {cause: "entityAttack", damagingEntity: player});
        }
        let equip = entity?.getComponent("minecraft:equippable");
        if ((equip?.getEquipment('Offhand')?.typeId === "minecraft:shield" || equip?.getEquipment('Mainhand')?.typeId === "minecraft:shield") && entity?.isSneaking) returnSignal = true;
    });
    return returnSignal;
}
function usedReduce(targetUseItem) {
    if (targetUseItem.getGameMode() !== "survival") {
        targetUseItem.amethystLock = undefined;
        return;
    }
    let currentItem = targetUseItem.getComponent("minecraft:equippable").getEquipment('Mainhand');
    let afterItem;
    if (currentItem?.amount > 1) {
        afterItem = new ItemStack(currentItem.typeId, currentItem.amount - 1);
    } else {
        afterItem = undefined;
    }
    targetUseItem.getComponent("minecraft:equippable").getEquipmentSlot('Mainhand').setItem(afterItem)
    targetUseItem.amethystLock = undefined;
}
function generateFarmland(main, player) {
    let downWardBlock = main.dimension.getBlockFromRay(main.location, {x: 0, y: -1, z: 0}, {maxDistance: 2, })?.block
    if (conversionDirtListHoe[downWardBlock?.typeId.replace('minecraft:', '')]) {
        if (downWardBlock.typeId === "minecraft:dirt_with_roots") player.dimension.spawnItem(new ItemStack("minecraft:hanging_roots"), downWardBlock.above().center());
        downWardBlock.setPermutation(BlockPermutation.resolve('minecraft:farmland').withState("moisturized_amount", 7));
    }
}
function generatePath(main, player) {
    let downWardBlock = main.dimension.getBlockFromRay(main.location, {x: 0, y: -1, z: 0}, {maxDistance: 2, })?.block
    if (conversionDirtListShovel[downWardBlock?.typeId.replace('minecraft:', '')]) {
        if (downWardBlock.typeId === "minecraft:dirt_with_roots") player.dimension.spawnItem(new ItemStack("minecraft:hanging_roots"), downWardBlock.above().center());
        downWardBlock.setPermutation(BlockPermutation.resolve('minecraft:grass_path'));
    }
}
function boomerangOut(player, entity, vec, dist, spawnItemStack, callFunction, coolDown = 0) {
    let returnSignal = false;
    if (coolDown === 0) {
        returnSignal = callFunction(entity, player, 15) //damageEntity(entity, player, 15)
    } else {
        coolDown -= 1;
    };
    if (returnSignal) coolDown = 2;
    if (dist > 15 || returnSignal) return boomerangIn(player, entity, spawnItemStack, callFunction);
    let loc = entity.location;
    if (!entity.tryTeleport({x: loc.x + vec.x, y: loc.y + vec.y, z: loc.z + vec.z}, {checkForBlocks: true})) return boomerangIn(player, entity, spawnItemStack, callFunction);
    system.runTimeout(() => boomerangOut(player, entity, vec, dist + 1, spawnItemStack, callFunction, coolDown), 1);
}
function boomerangIn(player, entity, spawnItemStack, callFunction) {
    callFunction(entity, player, 8) // damageEntity(entity, player, 8);
    let p = player.location, s = entity.location;
    p.y += 1.6;
    let dx = p.x - s.x, dy = p.y - s.y, dz = p.z - s.z;
    let d = Math.hypot(dx, dy, dz);
    if (d < 0.5) {
        entity.remove();
        let durability = spawnItemStack.getComponent("minecraft:durability");
        if (durability.damage > durability.maxDurability - 2) {
            world.playSound("random.break", player.location); 
            return;
        } else {
            durability.damage += 1;
            let emptyHand = player.getComponent("minecraft:equippable").getEquipmentSlot('Mainhand');
            if (emptyHand.getItem() === undefined) {emptyHand.setItem(spawnItemStack)} else {player.dimension.spawnItem(spawnItemStack, player.location)}
            return;
        }
    };
    let scale = 1 / d;
    entity.tryTeleport({x: s.x + dx * scale, y: s.y + dy * scale, z: s.z + dz * scale});
    system.runTimeout(() => boomerangIn(player, entity, spawnItemStack, callFunction), 1);
}
function* floodFillGenerator(startBlock, originalTypes, belowAllow, additionalPerm) {
    const queue = [[startBlock, 0]];
    const visitCount = new Map();
    const directions = ["north", "east", "south", "west", "above", ...(belowAllow ? ["below"] : [])];
    const MAX_VISITS = directions.length;
    while (queue.length > 0) {
        const [block, distance] = queue.shift();
        if (!block) continue;
        const locKey = JSON.stringify(block.location);
        const currentVisits = visitCount.get(locKey) || 0;
        if (currentVisits >= MAX_VISITS) continue;
        visitCount.set(locKey, currentVisits + 1);
        yield block;
        const neighbors = directions
            .map(dir => [block[dir](), distance + 1])
            .filter(([neighbor]) => neighbor && originalTypes.includes(neighbor.typeId))
            .filter(([neighbor]) => {
                const neighborKey = JSON.stringify(neighbor.location);
                const neighborVisits = visitCount.get(neighborKey) || 0;
                if (neighborVisits >= MAX_VISITS) return false;
                if (!additionalPerm) return true;
                if (neighbor.typeId.includes("_leaves") || neighbor.typeId.includes("wart_block")) {
                    return true;
                }
                return neighbor.permutation.getState(additionalPerm[0]) === additionalPerm[1];
            });
        const sortFn = (a, b) => {
            const aVisits = visitCount.get(JSON.stringify(a[0].location)) || 0;
            const bVisits = visitCount.get(JSON.stringify(b[0].location)) || 0;
            return a[1] - b[1] || aVisits - bVisits || Math.random() - 0.5;
        };
        
        queue.push(...neighbors.sort(sortFn));
        
        if (queue.length % 100 === 0) queue.sort(sortFn);
    }
}
function handleEntityConversion() {
    world.getAllPlayers().forEach(player => {
        let dim = player.dimension;
        dim.getEntities({families: ["item_entity_state"]}).forEach(entity => {
            dim.spawnItem(new ItemStack(entityConversionList[entity.typeId.replace("vi:", "")]), entity.location);
            entity.remove()
        });
    });
}
function dropVixiumInfused(block, blockType) {
    let vixiumRawAmount = vixiumRaw[blockType];
    if (vixiumRawAmount > 0) block.dimension.spawnItem(new ItemStack("vi:raw_vixium", vixiumRawAmount), block.center());
    block.dimension.spawnItem(new ItemStack("minecraft:amethyst_shard", shardAmount[blockType]), block.center());
};
world.afterEvents.playerBreakBlock.subscribe(e => {
    let block = e.block;
    if (e.brokenBlockPermutation.type.id === "vi:vixium_infused_amethyst_buds") {
        dropVixiumInfused(block, e.brokenBlockPermutation.getState("vi:state"));
    }
    if (e.brokenBlockPermutation.type.id === "vi:vixium_block") {
        ["north", "east", "west", "south", "above", "below"].forEach(dir => {
            let interBlock = block[dir]?.()
            if (interBlock?.typeId === "vi:vixium_infused_amethyst_buds" && dirToMethod[oppoDir[interBlock.permutation.getState("minecraft:facing_direction")]] === dir) {
                dropVixiumInfused(interBlock, interBlock.permutation.getState("vi:state"));
                interBlock.setPermutation(BlockPermutation.resolve('minecraft:air'));
            }
        });
    }
    if (e.itemStackAfterBreak?.typeId === "vi:vixium_pickaxe") {
        for (let item of block.dimension.getEntities({location: block.location, maxDistance: 2, type: "minecraft:item"})) {
            let itemOriginalName = item.getComponent("minecraft:item").itemStack.typeId
            if (itemOriginalName.includes("minecraft:raw_") && itemOriginalName.replace("minecraft:raw_", "") === e.brokenBlockPermutation.type.id.replace("minecraft:", "").replace("_ore", "").replace("deepslate_", "")) {
                item.remove();
                if (itemOriginalName !== "minecraft:raw_copper") break;
            }
        }
    }
})
world.beforeEvents.playerBreakBlock.subscribe(e => {
    let originalBlock = e.block.typeId;
    let startBlock = e.block;
    if (startBlock.typeId === "vi:vixium_infused_obsidian") {
        e.cancel = true;
        system.runTimeout(() => { 
            startBlock.dimension.spawnItem(new ItemStack("vi:raw_vixium", Math.floor(Math.random() * 2) + 1), startBlock.center());
            startBlock.setPermutation(BlockPermutation.resolve('minecraft:obsidian'));
        });
    };
    const toolMethod = {
        "vi:vixium_pickaxe": (originalBlock) => {
            if (!oreList.includes(originalBlock)) return;
            let searchArray = [];
            let redstoneList = ["minecraft:redstone_ore", "minecraft:lit_redstone_ore", "minecraft:deepslate_redstone_ore", "minecraft:lit_deepslate_redstone_ore"]
            if (redstoneList.includes(originalBlock)) {
                searchArray.push(...redstoneList)
            } else {
                searchArray.push(originalBlock)
            }
            let generator = floodFillGenerator(startBlock, searchArray, true);
            let loopSystem = system.runInterval(() => {
                let iterationCount = 0;
                let result = generator.next();
                while (!result.done && iterationCount < 8) {
                    let block = result.value;
                    if (searchArray.includes(block.typeId)) {
                        block.setPermutation(BlockPermutation.resolve('minecraft:air'));
                        let sound = 'dig.stone';
                        if (originalBlock.includes("deepslate")) {sound = 'dig.deepslate'} else if (originalBlock.includes("nether")) {sound = 'dig.netherrack'};
                        world.playSound(sound, block.location);
                        let dropAmount = Math.floor(Math.random() * 2) + 2;
                        let dropName = originalBlock.split(":")[1].replace("_ore", "").replace("deepslate_", "").replace("lit_", "");
                        if (oreMaxDrop[dropName]) dropAmount += 3;
                        e.player.dimension.spawnItem(new ItemStack(oreDropList[dropName], dropAmount), block.location);
                        block.dimension.spawnEntity("minecraft:xp_orb", block.center());
                        block.dimension.spawnEntity("minecraft:xp_orb", block.center());
                    }
                    result = generator.next();
                    iterationCount++;
                }
                if (result.done) {
                    system.clearRun(loopSystem);
                }
            }, 2);
            system.runTimeout(() => {system.clearRun(loopSystem)}, 80);
        },
        "vi:vixium_axe": (originalBlock) => {
            if (!e.player.isSneaking) return;
            let woodType = originalBlock.replace("minecraft:","").replace("_log", "").replace("_stem", "");
            if (woodConversionList[woodType] && startBlock.permutation.getState("pillar_axis") === "y") {
                let searchArray = [originalBlock, `minecraft:${woodType}_leaves`, `minecraft:${woodType}_stem`, "minecraft:warped_wart_block", "minecraft:nether_wart_block", "minecraft:shroomlight"];
                let generator = floodFillGenerator(startBlock, searchArray, false, ["pillar_axis", "y"]);
                let loopSystem = system.runInterval(() => {
                    let iterationCount = 0;
                    let result = generator.next();
                    while (!result.done && iterationCount < 10) {
                        let block = result.value;
                        block.dimension.runCommandAsync(`setblock ${block.location.x} ${block.location.y} ${block.location.z} air destroy`);
                        result = generator.next();
                        iterationCount++;
                    }
                    if (result.done) {
                        system.clearRun(loopSystem);
                    }
                }, 2);
                system.runTimeout(() => {system.clearRun(loopSystem);}, 60);
            }
        },
    }
    toolMethod[e.itemStack?.type.id]?.(originalBlock);
});
world.afterEvents.entityHitEntity.subscribe(e => {
    let hitEntity = e.hitEntity;
    if (hitEntity.typeId === "vi:nether_star_entity") {
        hitEntity.dimension.spawnItem(new ItemStack("minecraft:nether_star"), hitEntity.location);
        hitEntity.remove();
        netherStarLock = false;
    }
})
world.beforeEvents.itemUseOn.subscribe(e => {
    let block = e.block;
    let player = e.source;
    if (e.itemStack.typeId === "minecraft:amethyst_shard" && block.typeId === "vi:vixium_block") {
        let interBlock = block[dirToMethod[e.blockFace.toLowerCase()]]();
        if (interBlock?.typeId === "minecraft:air" && player.amethystLock === undefined) {
            player.amethystLock = true;
            system.runTimeout(() => {
                usedReduce(player)
                world.playSound("place.amethyst_cluster", block.location);
                interBlock.setPermutation(BlockPermutation.resolve('vi:vixium_infused_amethyst_buds').withState("vi:state", "small").withState("minecraft:facing_direction", oppoDir[e.blockFace.toLowerCase()]));
            });
        }
    };
});
world.afterEvents.itemUse.subscribe(e => {
    let player = e.source;
    let block = player.getBlockFromViewDirection()?.block;
    let dim = block?.dimension;
    if (e.itemStack.typeId === "minecraft:nether_star" && block?.typeId === "minecraft:lightning_rod" && !netherStarLock && dim?.id === "minecraft:overworld") {
        netherStarLock = true;
        let currentItem = player.getComponent("minecraft:inventory")?.container?.getItem(player?.selectedSlotIndex);
        let afterItem;
        if (currentItem.amount > 1) {
            afterItem = new ItemStack(currentItem.typeId, currentItem.amount - 1);
        } else {
            afterItem = undefined;
        }
        player.getComponent("minecraft:equippable").getEquipmentSlot('Mainhand').setItem(afterItem)
        dim.playSound("beacon.power", block.location);
        let netherStar = dim.spawnEntity("vi:nether_star_entity", block.above().center())
        system.runTimeout(() => {
            if (netherStar.isValid()) {
                dim.spawnEntity("minecraft:lightning_bolt", netherStar.location);
                dim.playSound("beacon.ambient", block.location);
            }
        }, 60);
        system.runTimeout(() => {
            if (netherStar.isValid()) {
                dim.spawnEntity("minecraft:lightning_bolt", netherStar.location)
            }
        }, 120);
        system.runTimeout(() => {
            if (netherStar.isValid()) {
                dim.playSound("beacon.deactivate", block.location);
                dim.spawnEntity("minecraft:lightning_bolt", netherStar.location)
            }
        }, 180);
        system.runTimeout(() => {
            if (netherStar.isValid()) {
                dim.spawnParticle("vi:meteor_large", block.center());
                world.playSound("sound.vixium_landing", block.location);
                netherStar.remove();
                system.runTimeout(() => {
                    netherStarLock = false
                    dim.spawnParticle("vi:explosion_melt_large", block.center());
                    world.playSound("sound.vixium_impact", block.location);
                    dim.getEntities({location: block.location, maxDistance: 25, type: "minecraft:player"}).forEach(player => {
                        dim.createExplosion(block.location, 12, {causesFire: true})
                        player.applyDamage(10 + 150/(Math.sqrt((player.location.x - block.location.x) ** 2 + (player.location.z - block.location.z) ** 2) + 1), {cause: "blockExplosion"})
                        world.structureManager.place(`mystructure:asteriod_${Math.floor(Math.random() * 4) + 1}`, dim, dim.getBlockFromRay(block.north(2).west(2).location, {x: 0, y: -1, z: 0}).block.location, {rotation: Math.floor(Math.random() * 4) * 90});
                    });
                }, 200);
            }
        }, 240);
    }
});
world.beforeEvents.worldInitialize.subscribe(e => {
    e.blockComponentRegistry.registerCustomComponent('vi:vixium_infused_amethyst_buds', {
        onTick: (e) => {
            let block = e.block;
            if (block.permutation.getState("vi:state") === "large") {
                block.dimension.spawnParticle("vi:twinkling_vixium", block.center())
            };
            if (Math.random() < 0.01) {
                let newState = stateGrowth[block.permutation.getState("vi:state")];
                if (newState) {
                    block.setPermutation(block.permutation.withState("vi:state", newState));
                }
            }
        }
    });
    e.blockComponentRegistry.registerCustomComponent('vi:vixium_infused_obsidian', {
        onTick: (e) => {
            e.block.dimension.spawnParticle("vi:twinkling_vixium", e.block.center())
        }
    });
    e.itemComponentRegistry.registerCustomComponent('vi:vixium_axe', {
        onUseOn: (e) => {
            let block = e.block;
            const oldStates = e.usedOnBlockPermutation;
            let iBlock = oldStates.type.id;
            if (iBlock.includes("waxed")) {
                let newBlockName;
                if (iBlock.includes("_door")) {
                    if (block.permutation.getState("upper_block_bit")) block = block.below();
                    let loc = block.location;
                    block.dimension.runCommand(`setblock ${loc.x} ${loc.y} ${loc.z} ${iBlock.replace("waxed_", "").replace("minecraft:", "")} ["direction"=${oldStates.getState("direction")},"door_hinge_bit"=${oldStates.getState("door_hinge_bit")},"open_bit"=${oldStates.getState("open_bit")}]`)
                } else {
                    const blockTypes = ["bulb", "trapdoor", "slab", "stairs", "chiseled"];
                    if (blockTypes.some(type => iBlock.includes(type))) {
                        newBlockName = iBlock.replace("waxed_", "");                        
                    } else if (iBlock.includes("grate")) {
                        newBlockName = iBlock.replace("waxed_", "").replace("_block", "")
                    } else {
                        newBlockName = iBlock.replace("waxed_", "") + "_block"
                    };
                    let newPerm = BlockPermutation.resolve(newBlockName);
                    for (const state in oldStates.getAllStates()) {
                        newPerm = newPerm.withState(state, oldStates.getState(state));
                    }
                    if (iBlock.includes("door")) block.above().setPermutation(BlockPermutation.resolve("minecraft:air"));
                    block.setPermutation(newPerm);
                };
                block.dimension.spawnParticle("vi:waxed_off", block.center())
                world.playSound("copper.wax.off", block.location);
            };
            if (!iBlock.startsWith("minecraft:stripped")) {
                if (iBlock === "minecraft:bamboo_block") {
                    world.playSound("break.bamboo_wood", block.location);
                    block.setPermutation(BlockPermutation.resolve(iBlock.replace("minecraft:", "minecraft:stripped_")).withState("pillar_axis", oldStates.getState("pillar_axis")))
                };
                if (iBlock.includes("log") || iBlock.includes("wood") || iBlock.endsWith("_stem")) {
                    world.playSound("use.wood", block.location);
                    block.setPermutation(BlockPermutation.resolve(iBlock.replace("minecraft:", "minecraft:stripped_")).withState("pillar_axis", oldStates.getState("pillar_axis")));
                };
                if (iBlock.includes("_hyphae")) {
                    world.playSound("step.nether_wood", block.location);
                    block.setPermutation(BlockPermutation.resolve(iBlock.replace("minecraft:", "minecraft:stripped_")));
                };
            };
        },
        onHitEntity: (e) => {
            if (e.hitEntity.isSneaking) {
                let equip = e.hitEntity.getComponent("minecraft:equippable");
                let itemOff = equip.getEquipment('Offhand');
                if (itemOff?.typeId === "minecraft:shield") {
                    itemOff.getComponent('minecraft:cooldown').startCooldown(e.hitEntity)
                };
                let itemMain = equip.getEquipment('Mainhand');
                if (itemMain?.typeId === "minecraft:shield") {
                    itemMain.getComponent('minecraft:cooldown').startCooldown(e.hitEntity)
                };
            }
        }
    });
    e.itemComponentRegistry.registerCustomComponent('vi:vixium_shovel', {
        onUseOn: (e) => {
            let interactedBlock = e.usedOnBlockPermutation.type.id.replace("minecraft:", "");
            let turnTo = conversionDirtListShovel[interactedBlock]
            if (interactedBlock === "dirt_with_roots") {
                e.source.dimension.spawnItem(new ItemStack("minecraft:hanging_roots"), e.block.center());
            };
            if (turnTo) {
                e.block.setPermutation(BlockPermutation.resolve('minecraft:grass_path'));
            };
            if (e.usedOnBlockPermutation?.getState("extinguished") === false) {
                world.playSound("random.fizz", e.block.location, {volume: 0.5, pitch: 1.5});
                e.block.setPermutation(BlockPermutation.resolve('minecraft:' + interactedBlock).withState("extinguished", true).withState("minecraft:cardinal_direction", e.usedOnBlockPermutation.getState("minecraft:cardinal_direction")));
            }
        },
        onUse: (e) => {
            let player = e.source;
            if (!player.getBlockFromViewDirection({maxDistance: 7})) {
                player.getComponent("minecraft:equippable").getEquipmentSlot('Mainhand').setItem(undefined)
                let loc = player.location;
                let vixiumShovel = player.dimension.spawnEntity('vi:vixium_shovel_throwing', {x: loc.x, y: loc.y+0.5, z: loc.z});
                let moveDir = player.getViewDirection();
                moveDir.y = 0;
                boomerangOut(player, vixiumShovel, moveDir, 8, e.itemStack, generatePath);
            }
        }
    });
    e.itemComponentRegistry.registerCustomComponent('vi:vixium_sword', {
        onHitEntity: (e) => {
            let target = e.hitEntity;
            let attackingEntity = e.attackingEntity;
            target?.dimension.getEntities({location: target.location, maxDistance: 1.75}).forEach(entity => {
                if (entity !== attackingEntity) {
                    entity.applyDamage(5, {cause: "entityAttack", damagingEntity: attackingEntity});
                }
            });
        },
        onUse: (e) => {
            let player = e.source;
            player.getComponent("minecraft:equippable").getEquipmentSlot('Mainhand').setItem(undefined)
            let loc = player.location;
            let vixiumSword = player.dimension.spawnEntity('vi:vixium_sword_throwing', {x: loc.x, y: loc.y+1.6, z: loc.z});
            boomerangOut(player, vixiumSword, player.getViewDirection(), 0, e.itemStack, damageEntity, 0);
        }
    });
    e.itemComponentRegistry.registerCustomComponent('vi:vixium_trident', {
        onCompleteUse: (e) => {
            let player = e.source;
            player.playSound("sound.vixium_trident");
            const { x, y, z } =  player.getViewDirection();
            const horizontalStrength = 7 * Math.sqrt(x * x + z * z);
            player.applyKnockback(
                horizontalStrength !== 0 ? x / horizontalStrength : 0,
                horizontalStrength !== 0 ? z / horizontalStrength : 0,
                horizontalStrength,
                2 * y + 0.2* +player.isOnGround,
            );
            player.addEffect("slow_falling", 40, {showParticles: false});
            const curSlot = player.selectedSlotIndex;
            const cont = player.getComponent("minecraft:inventory")?.container;
            cont.setItem(curSlot, new ItemStack("vi:vixium_trident_cooldown"));
            let tridentCooldown = cont.getItem(curSlot);
            tridentCooldown.getComponent('minecraft:cooldown').startCooldown(player);
            let damageLoop = system.runInterval(() => {
                let dir = player.getViewDirection();
                let loc = player.location;
                let front = {x: loc.x + 2 * dir.x, y: 0.75 + loc.y + 2 * dir.y, z: loc.z + 2 * dir.z};
                player.dimension.spawnParticle("vi:spark_vixium", {x: loc.x + 8 * dir.x, y: 0.75 + loc.y + 8 * dir.y, z: loc.z + 8 * dir.z});
                player.dimension.getEntities({location: front, maxDistance: 3}).forEach(entity => {
                    let family = entity.getComponent("minecraft:type_family");
                    if ((family?.hasTypeFamily("monster") || family?.hasTypeFamily("player")) && entity !== player) {
                        player.applyKnockback(
                            horizontalStrength !== 0 ? -(x / horizontalStrength) : 0,
                            horizontalStrength !== 0 ? -(z / horizontalStrength) : 0,
                            0.85 * horizontalStrength,
                            -0.1 * y,
                        );
                        if (entity.typeId !== "minecraft:player") entity.applyImpulse({x: 2.5 * dir.x, y: 2.5 * dir.y, z: 2.5 * dir.z});
                        entity.applyDamage(12, {cause: "entityAttack", damagingEntity: player});
                        system.clearRun(damageLoop);
                    }
                });
                if (player.isOnGround) system.clearRun(damageLoop);
            });
            system.runTimeout(() => {
                system.clearRun(damageLoop);
                for (let i = 0; i < cont.size; i++) {
                    if (cont.getItem(i)?.typeId === "vi:vixium_trident_cooldown") {
                        cont.setItem(i, new ItemStack("vi:vixium_trident"));
                    }
                }
            }, 10);
        }
    });
    e.itemComponentRegistry.registerCustomComponent('vi:vixium_trident_cooldown', {
        onUse: (e) => {
            let player = e.source;
            player.getComponent("minecraft:inventory")?.container?.setItem(player.selectedSlotIndex, new ItemStack("vi:vixium_trident"));
        }
    });
    e.itemComponentRegistry.registerCustomComponent('vi:vixium_hoe', {
        onUseOn: (e) => {
            let interactedBlock = e.usedOnBlockPermutation.type.id.replace("minecraft:", "");
            let turnTo = conversionDirtListHoe[interactedBlock];
            if (interactedBlock === "dirt_with_roots") {
                e.source.dimension.spawnItem(new ItemStack("minecraft:hanging_roots"), e.block.center());
                turnTo = "dirt"
            };
            if (turnTo) {
                let permutationToSet = BlockPermutation.resolve(`minecraft:${turnTo}`);
                if (turnTo === "farmland") permutationToSet = BlockPermutation.resolve(`minecraft:${turnTo}`).withState("moisturized_amount", 7);
                e.block.setPermutation(permutationToSet);
            }
        },
        onUse: (e) => {
            let player = e.source;
            if (!player.getBlockFromViewDirection({maxDistance: 7})) {
                player.getComponent("minecraft:equippable").getEquipmentSlot('Mainhand').setItem(undefined)
                let loc = player.location;
                let vixiumHoe = player.dimension.spawnEntity('vi:vixium_hoe_throwing', {x: loc.x, y: loc.y+0.5, z: loc.z});
                let moveDir = player.getViewDirection();
                moveDir.y = 0;
                boomerangOut(player, vixiumHoe, moveDir, 4, e.itemStack, generateFarmland);
            }
        }
    });
});
world.afterEvents.playerJoin.subscribe(e => {
    system.runTimeout(() => handleEntityConversion(), 100);
});
handleEntityConversion()